#include "music.h"
#include <vector>
#include <string.h>
#include <math.h>
#include <algorithm>
#include <chrono>
#include <fstream>
#include <iostream>
#include <random>


using namespace std;
typedef chrono::high_resolution_clock Time;
typedef chrono::milliseconds ms;
typedef chrono::duration<float> fsec;

void print_help() {
    cout << "Well-formed input file is as follows:" << endl;
    cout << "generate num_sources <number_of_sources> num_antennas " 
         << "<number_of_antennas> angles <angles_in_deg> <outfile_name>" << endl;

    cout << "or" << endl;
    cout << "covmat num_sources <number_of_sources> num_antennas " 
         << "<number_of_antennas> angles <angles_in_deg> "
         << "<covariance_matrix_elements>" << endl;

    cout << "Angles should fall between 0 and 180 degrees." << endl;
    cout << "outfile_name of 'none will produce no output file." << endl;
    cout << "The covariance matrix should be of size <number_of_antennas> x "
         << "<number_of_antennas>." << endl;
    cout << "The covariance matrix should be a space saperated list of complex "
         << "doubles. Example in cov.in" << endl;

}

// Use thetas to generate 100 relazations and form a covariance matrix. Calls
// The MUSIC algorithm with the covariance matrix to get estimated angles.
vector<double> RunMusic(vector<double>& Thetas, int N, string filename) {
    int rand_seed = 6;
    default_random_engine generator;
    normal_distribution<double> distribution(0,1);
    srand(rand_seed);
    int L = Thetas.size();

    complex<double> snr(10,0);
    double pi = 3.14159;
    vector<complex<double>> array(N);

    double step = ((double) N ) / (2 * (N));
    for (int i = 0; i < N; i++) {
        array[i] = step*i;
    } 

    // Generate Angles which has 360 angles between -pi/2 and pi/2.
    vector<double> Angles(360);
    step = pi /359;
    for (int i = 0 ; i < 360 ; i++) {
        Angles[i] = step* i - pi/2;
    }

    int numrealization = 100;
    Eigen::MatrixXcd H(N,numrealization);

    // Generate true directions of arrival
    vector<complex<double>> Alphas(L);
    rand();
    for (int i = 0; i < L; i++) {
        Thetas[i] -= 90; // Shift 90 degrees.
        Thetas[i] *= 0.0174533; // Convert deg to rad.
        Alphas[i] = complex<double>(sqrt(.5)*(distribution(generator)),
                                    sqrt(.5)*(distribution(generator)));
    }

    // Generates vectors that will be used to form the covariance matrix.
    for (int iter = 0; iter < numrealization; iter++) {
        Eigen::VectorXcd htmp(N);
        for (int i = 0; i < N; i++) {
            htmp[i] = 0;
        }
        for (int i = 0; i < L; i++) {
            complex<double> t = complex<double>(0,1);

            complex<double> pha = exp(t*complex<double>(2*pi,0) * 
                                      complex<double>(((double) rand() / (RAND_MAX)) , 0));
            htmp = htmp + pha*Alphas[i]*array_response_vector(array,Thetas[i]);
        }
        for (int i = 0; i < N; i ++) {
            htmp[i] = htmp[i] + sqrt(.5/snr) * 
                      complex<double>(distribution(generator),
                                      distribution(generator));
        }
        H.col(iter) = htmp;
    }

    // Covariance matrix.
    Eigen::MatrixXcd CovMat = (H * H.conjugate().transpose());
    ofstream outfile;

    if (filename != "none") {
        outfile.open(filename);
        outfile << "covmat\n";
        outfile << "num_sources " << Thetas.size() << "\n";
        outfile << "num_antennas " << N << "\n";
        outfile << "angles ";
        for (int i = 0; i < Thetas.size(); i++) {
            outfile << (Thetas[i] / 0.0174533 + 90) << " ";
        }
        outfile << "\n";
        for (int i = 0; i < CovMat.cols(); i++) {
            for (int j = 0; j < CovMat.cols(); j++) {
                outfile << CovMat.row(i)[j].real() << " ";
                outfile << CovMat.row(i)[j].imag() << " ";
            }
        }

    }


    // MUSIC
    vector<double> estimated = music(CovMat, L, N, array, Angles);

    // Sort the true angles from smallest to largest.
    std::sort(Thetas.begin(), Thetas.end());
    for (int i = 0; i < Thetas.size(); i ++ ){
      // Convert angles back to degrees and shift back 90 degrees.
      estimated[i] = (estimated[i]/ .0174533) + 90;
      Thetas[i] = (Thetas[i] / 0.0174533 + 90);
    }

    return estimated;
}

// Use thetas to generate 100 relazations and form a covariance matrix. Calls
// The MUSIC algorithm with the covariance matrix to get estimated angles.
vector<double> RunMusic(Eigen::MatrixXcd& CovMat, vector<double>& Thetas, int N) {
    complex<double> snr(10,0);
    double pi = 3.14159;
    vector<complex<double>> array(N);

    double step = ((double) N ) / (2 * (N));
    for (int i = 0; i < N; i++) {
        array[i] = step*i;
    } 

    // Generate Angles which has 360 angles between -pi/2 and pi/2.
    vector<double> Angles(360);
    step = pi /359;
    for (int i = 0 ; i < 360 ; i++) {
        Angles[i] = step* i - pi/2;
    }

    // MUSIC
    vector<double> estimated = music(CovMat, Thetas.size(), N, array, Angles);

    // Sort the true angles from smallest to largest.
    std::sort(Thetas.begin(), Thetas.end());
    for (int i = 0; i < Thetas.size(); i ++ ){
      // Convert angles back to degrees and shift back 90 degrees.
      estimated[i] = (estimated[i]/ .0174533) + 90;
    }

    return estimated;
}

int main(int argc, char** argv) {
    if (argc != 2) {
        cerr << "Usage ./music [input filename]\n";
        return 0;
    }

    if (!strcmp(argv[1], "--help")) {
        print_help();
        return 0;
    }

    ifstream infile(argv[1]);

    if (!infile.is_open()) {
        cerr << "Input file cannot be found.\n";
        return 1;
    }

    string method;
    infile >> method;
    // Get input params.
    string temp;
    infile >> temp;
    if (temp != "num_sources") {
        print_help();
        return 1;
    }

    infile >> temp;
    int L = stoi(temp);

    infile >> temp;
    if (temp != "num_antennas") {
        print_help();
        return 1;
    }
    infile >> temp;
    int N = stoi(temp);

    if (L >= N) {
        cout << "Number of antennas must exceed number of sources" << endl;
        return 1;
    }

    infile >> temp;
    if (temp != "angles") {
        print_help();
        return 1;
    }
    vector<double> angles(L);

    for (int i = 0; i < L; i++) {
        infile >> temp;
        angles[i] = stod(temp);
    }


    cout << "Running MUSIC algorithm for DoA with " << L << " sources and "
         << N << " antennas" << "\n";  

    // Run MUSIC. 
    auto time_state = Time::now(); 
    vector<double> estimated_angles;
    if (method == "generate") {
        infile >> temp;
        estimated_angles = RunMusic(angles,N,temp);
    } else if (method == "covmat") {
        Eigen::MatrixXcd CovMat(N,N);
        string temp2;
        for (int i = 0; i < N; i++) {
            for (int j = 0; j < N; j++) {
                infile >> temp;
                infile >> temp2;
                CovMat.row(i)[j] = complex<double>(stod(temp),stod(temp2));
            }
        }
        time_state = Time::now();
        estimated_angles = RunMusic(CovMat,angles,N);
    } else {
        print_help();
        return 1;
    }
    fsec total_time = (Time::now() - time_state);

    // Print results.
    cout << "Estimated angles: \n";
    for (int i = 0; i < estimated_angles.size(); i++) {
      cout << estimated_angles[i]<< " ";
    }
    cout << "\n";

    cout << "Actual angles: \n";
    for (int i = 0; i < angles.size(); i++) {
      cout << angles[i]<< " ";
    }
    cout << "\n";

    double mse = 0;
    double avg_diff = 0;
    double max_diff = 0;
    for (int i = 0; i < angles.size(); i++) {
      mse += (angles[i] - estimated_angles[i])*(angles[i] - estimated_angles[i]);
      avg_diff += abs(angles[i] - estimated_angles[i]);
      if (abs(angles[i] - estimated_angles[i]) > max_diff) {
        max_diff = abs(angles[i] - estimated_angles[i]);
      }
    } 
    mse = mse / angles.size();
    avg_diff = avg_diff/angles.size();
    cout << "Mean squared error: " << mse << "\n";
    cout << "Average angle difference: " << avg_diff << " degrees\n";
    cout << "Largest angle difference: " << max_diff << " degrees\n";
    cout << "Total time taken: " << total_time.count() << "s\n";  

    return 0;
}
